//-----------------------------------------------------------------
// TicTacToe Game Object
// C++ Source - Game.cpp
//-----------------------------------------------------------------

//-----------------------------------------------------------------
// Inclusions
//-----------------------------------------------------------------
#include "StdAfx.h"
//-----------------------------------------------------------------
#include "Game.h"
#include "ServSDlg.h"
#include "SessIDlg.h"
#include "SessSDlg.h"
#include "PlayIDlg.h"
#include "PlaySDlg.h"
#include "Wave.h"
#include "GUID.h"
#include "Resource.h"

//-----------------------------------------------------------------
// MFC Debugging Support
//-----------------------------------------------------------------
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif


//-----------------------------------------------------------------
// CGame Public Constructor(s)/Destructor
//-----------------------------------------------------------------
CGame::CGame()
  : m_pDirectPlay(NULL), m_hDPEvent(NULL), m_pEventThread(NULL),
    m_bMyTurn(FALSE)
{
  // Clear the game grid
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      m_tsGrid[i][j] = tsEMPTY;

  DPInit();
}

CGame::~CGame()
{
  DPCleanup();
}

//-----------------------------------------------------------------
// CGame Public Methods
//-----------------------------------------------------------------
BOOL
CGame::DPInit()
{
  // Clear the players
  m_dpidPlayer[0] = 0;
  m_dpidPlayer[1] = 0;

  // Prompt user to select a DP server, then create the DP object
  CServerSelDlg dlgServerSel;
  if (dlgServerSel.DoModal() == IDOK)
    return (::DirectPlayCreate(dlgServerSel.GetSelServer(),
      &m_pDirectPlay, NULL) == DP_OK);
    
  return FALSE;
}

void
CGame::DPCleanup()
{
  // End the DP session
  DPEndSession();

  // Cleanup the DP object
  if (m_pDirectPlay)
  {
    m_pDirectPlay->Release();
    m_pDirectPlay = NULL;
  }
}

BOOL
CGame::DPCreateSession()
{
  if (m_pDirectPlay)
  {
    // Get session information
    CSessionInfoDlg dlgSessionInfo;
    if (dlgSessionInfo.DoModal() == IDOK)
    {
      // Create a new DP session
      DPSESSIONDESC dpsdDesc;
      ::ZeroMemory(&dpsdDesc, sizeof(DPSESSIONDESC));
      dpsdDesc.dwSize = sizeof(DPSESSIONDESC);
      dpsdDesc.dwMaxPlayers = 2;
      dpsdDesc.dwFlags = DPOPEN_CREATESESSION;
      dpsdDesc.guidSession = TICTACTOE_10;
      ::strcpy(dpsdDesc.szSessionName, dlgSessionInfo.GetName());
      if (m_pDirectPlay->Open(&dpsdDesc) == DP_OK)
      {
        // Create local player and set game info
        m_pDirectPlay->EnableNewPlayers(TRUE);
        if (DPCreateLocalPlayer())
        {
          DPCreateEventThread();
          m_bMyTurn = TRUE;
          return TRUE;
        }
      }
    }
  }

  return FALSE;
}

BOOL
CGame::DPConnectSession()
{
  if (m_pDirectPlay)
  {
    // Select a DP session
    CSessionSelDlg dlgSessionSel(m_pDirectPlay);
    if (dlgSessionSel.DoModal() == IDOK)
    {
      // Open remote DP session
      DPSESSIONDESC dpsdDesc;
      ::ZeroMemory(&dpsdDesc, sizeof(DPSESSIONDESC));
      dpsdDesc.dwSize = sizeof(DPSESSIONDESC);
      dpsdDesc.dwFlags = DPOPEN_OPENSESSION;
      dpsdDesc.guidSession = TICTACTOE_10;
      dpsdDesc.dwSession = dlgSessionSel.GetSelSession();
      if (m_pDirectPlay->Open(&dpsdDesc) == DP_OK)
      {
        // Prompt user to select the remote player
        CPlayerSelDlg dlgPlayerSel(m_pDirectPlay);
        if (dlgPlayerSel.DoModal() == IDOK)
        {
          // Set remote player
          m_dpidPlayer[1] = dlgPlayerSel.GetSelPlayer();

          // Create local player and set game info
          m_pDirectPlay->EnableNewPlayers(TRUE);
          if (DPCreateLocalPlayer())
          {
            DPCreateEventThread();
            m_bMyTurn = FALSE;
            NewGame();
            return TRUE;
          }
        }
      }
    }
  }

  return FALSE;
}

void
CGame::DPEndSession()
{
  // Destroy the local DP player and clear the remote player
  DPDestroyLocalPlayer();
  m_dpidPlayer[1] = 0;

  // Close the DP object
  if (m_pDirectPlay)
    m_pDirectPlay->Close();

  // Cleanup the event thread
  if (m_pEventThread)
  {
    delete m_pEventThread;
    m_pEventThread = NULL;
  }
}

BOOL
CGame::NewGame()
{
  if (!DPIsValid())
    return FALSE;

  // Empty the grid
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      m_tsGrid[i][j] = tsEMPTY;

  // Reset turn count
  m_uiTurns = 0;

  // Update grid
  AfxGetMainWnd()->Invalidate(TRUE);

  // Notify of new game
  if (m_bMyTurn)
    AfxGetMainWnd()->MessageBox("New game, it's your turn.",
      AfxGetAppName());
  else
  {
    // Get the remote player's name
    char  szFriendlyName[DPSHORTNAMELEN];
    DWORD dwNameLen = DPSHORTNAMELEN;
    if (m_pDirectPlay->GetPlayerName(m_dpidPlayer[1],
      szFriendlyName, &dwNameLen, NULL, NULL) == DP_OK)
    {
      CString sText;
      sText.Format("New game, it's %s's turn.", szFriendlyName);
      AfxGetMainWnd()->MessageBox(sText, AfxGetAppName());
    }
  }

  return TRUE;
}

BOOL
CGame::SetTileState(UINT uiX, UINT uiY)
{
  ASSERT((uiX < 3) && (uiY < 3));

  CWave wavTile;
  if (m_tsGrid[uiX][uiY] == tsEMPTY)
  {
    // Send tile info to remote player via a turn message
    if (m_bMyTurn)
      if (!DPSendTurnMsg(CPoint(uiX, uiY)))
      {
        AfxGetMainWnd()->MessageBox("Error sending turn message.",
          AfxGetAppName());
        return FALSE;
      }

    // Change turns and set the tile state
    m_bMyTurn = !m_bMyTurn;
    m_tsGrid[uiX][uiY] = (m_uiTurns % 2) ? tsO : tsX;

    // Update grid
    AfxGetMainWnd()->Invalidate(FALSE);

    // Play the tile wave
    wavTile.Create((m_uiTurns % 2) ? IDW_O : IDW_X);
    wavTile.Play();
  }
  else
  {
    // Play the tile error wave
    wavTile.Create(IDW_ERROR);
    wavTile.Play();
    return FALSE;
  }

  // Check for winner/draw
  if (IsWinner())
  {
    // Determine winner and notify
    if (m_bMyTurn)
    {
      CWave wavLose(IDW_LOSE);
      wavLose.Play();
      AfxGetMainWnd()->MessageBox("Bummer, you lost!",
        AfxGetAppName());
    }
    else
    {
      CWave wavWin(IDW_WIN);
      wavWin.Play();
      AfxGetMainWnd()->MessageBox("Congratulations, you won!",
        AfxGetAppName());
    }

    // Start new game
    return NewGame();
  }
  else
  {
    if (IsDraw())
    {
      // Play draw wave
      CWave wavDraw(IDW_DRAW);
      wavDraw.Play();

      // Notify of a draw
      AfxGetMainWnd()->MessageBox("It's a draw!", AfxGetAppName());

      // Start new game
      return NewGame();
    }
  }

  return TRUE;
}

//-----------------------------------------------------------------
// CGame Protected Methods
//-----------------------------------------------------------------
BOOL
CGame::DPCreateLocalPlayer()
{
  // Create local DP player
  CPlayerInfoDlg dlgPlayerInfo;
  if (dlgPlayerInfo.DoModal() == IDOK)
    return (m_pDirectPlay->CreatePlayer(&m_dpidPlayer[0],
      dlgPlayerInfo.GetFriendlyName(),
      dlgPlayerInfo.GetFormalName(), &m_hDPEvent) == DP_OK);

  return FALSE;
}

BOOL
CGame::DPDestroyLocalPlayer()
{
  // Destroy local DP player
  if (m_pDirectPlay->DestroyPlayer(m_dpidPlayer[0]) == DP_OK)
  {
    m_dpidPlayer[0] = 0;
    return TRUE;
  }

  return FALSE;
}

BOOL
CGame::DPSendTurnMsg(CPoint& ptTile)
{
  // Send turn message to remote player
  if (m_pDirectPlay->Send(m_dpidPlayer[0], m_dpidPlayer[1],
    DPSEND_GUARANTEE, (LPVOID)(POINT*)&ptTile, sizeof(POINT)) ==
    DP_OK)
    return TRUE;

  return FALSE;
}

BOOL
CGame::DPReceiveTurnMsg(CPoint& ptTile)
{
  // Check remote turn message for valid tile bounds
  if ((ptTile.x >= 0) && (ptTile.x <= 2) && (ptTile.y >= 0) &&
    (ptTile.y <= 2))
  {
    // Update game with remote turn data
    SetTileState(ptTile.x, ptTile.y);

    return TRUE;
  }

  return FALSE;
}

void
CGame::DPCreateEventThread()
{
  // Cleanup existing thread, if necessary
  if (m_pEventThread)
    delete m_pEventThread;

  m_pEventThread = AfxBeginThread(DPEventMsgStart, (LPVOID)this);
}

UINT
CGame::DPEventMsgStart(LPVOID pData)
{
  // Call the DP event handler
  ASSERT((CGame*)pData);
  ((CGame*)pData)->DPEventMsg();

  return 0;
}

void
CGame::DPEventMsg()
{
  while(TRUE)
  {
    // Wait for event
    if (::WaitForSingleObject(m_hDPEvent, INFINITE) !=
      WAIT_TIMEOUT)
    {
      // Process event message
      if (m_pDirectPlay)
      {
        DPID  dpidFrom, dpidTo;
        BYTE  Msg[256];
        DWORD dwLen = 128;

        if (m_pDirectPlay->Receive(&dpidFrom, &dpidTo,
          DPRECEIVE_ALL, Msg, &dwLen) == DP_OK)
        {
          if (dpidFrom == 0)
          {
            // Got a system message
            DPMSG_GENERIC*  pmsgGeneric = (DPMSG_GENERIC*)Msg;
            CString         sText;
            switch(pmsgGeneric->dwType)
            {
              case DPSYS_CONNECT:
                AfxGetMainWnd()->MessageBox("Connected!",
                  AfxGetAppName());
                break;

              case DPSYS_SESSIONLOST:
                AfxGetMainWnd()->MessageBox("Session lost!",
                  AfxGetAppName());
                DPCleanup();
                break;

              case DPSYS_ADDPLAYER:
                // Notify of new player
                sText.Format("New Player : %s", ((DPMSG_ADDPLAYER*)
                  pmsgGeneric)->szShortName);
                AfxGetMainWnd()->MessageBox(sText,
                  AfxGetAppName());

                // Set new player and start new game
                if (((DPMSG_ADDPLAYER*)pmsgGeneric)->dpId !=
                  m_dpidPlayer[0])
                {
                  m_dpidPlayer[1] = ((DPMSG_ADDPLAYER*)
                    pmsgGeneric)->dpId;
                  NewGame();
                }
                break;

              case DPSYS_DELETEPLAYER:
                AfxGetMainWnd()->MessageBox("Player Deleted!",
                  AfxGetAppName());
                if (((DPMSG_DELETEPLAYER*)pmsgGeneric)->dpId ==
                  m_dpidPlayer[1])
                {
                  m_dpidPlayer[1] = 0;
                  DPEndSession();
                }
                break;
            }
          }
          else
            if (dpidTo == m_dpidPlayer[0])
            {
              // Got a remote player turn message
              if (dwLen == sizeof(POINT))
              {
                CPoint ptTile(*((POINT*)Msg));
                DPReceiveTurnMsg(ptTile);
              }
              else
                AfxGetMainWnd()->MessageBox(
                  "Unknown player message.", AfxGetAppName());
            }
        }
      }
    }
  }
}

BOOL
CGame::CheckLine(int i, int j, int ii, int jj)
{
  TileState ts = m_tsGrid[i][j];

  if (ts == tsEMPTY)
    return FALSE;

  // Check for unequal tiles along line
  for (int k = 0; k < 2; k++)
    if (ts != m_tsGrid[i += ii][j += jj])
      return FALSE;

  return TRUE;
}

BOOL
CGame::IsWinner()
{
  // Check for row/column winner
  for (int i = 0; i < 3; i++)
  {
    // Check rows
    if (CheckLine(i, 0, 0, 1))
      return TRUE;

    // Check columns
    if (CheckLine(0, i, 1, 0))
      return TRUE;
  }

  // Check for diagonal winner
  if (CheckLine(0, 0, 1, 1) || CheckLine(0, 2, 1, -1))
    return TRUE;

  return FALSE;
}

BOOL
CGame::IsDraw()
{
  // See if game is a draw
  if (++m_uiTurns < 9)
    return FALSE;

  return TRUE;
}
