//-----------------------------------------------------------------
// Game Object
// C++ Header - Game.h
//-----------------------------------------------------------------

#ifndef __GAME_H__
#define __GAME_H__

//-----------------------------------------------------------------
// Inclusions
//-----------------------------------------------------------------
#include <DPlay.h>

//-----------------------------------------------------------------
// CGame Class - Game Object
//-----------------------------------------------------------------
class CGame : public CObject
{
  // Public Types
public:
  enum TileState
  {
    tsEMPTY,
    tsX,
    tsO
  };

  // Public Constructor(s)/Destructor
public:
          CGame();
  virtual ~CGame();

  // Public Methods
public:
  BOOL                DPInit();
  void                DPCleanup();
  BOOL                DPCreateSession();
  BOOL                DPConnectSession();
  void                DPEndSession();
  BOOL                DPIsValid() const { return (m_dpidPlayer[0]
                        && m_dpidPlayer[1]); }
  BOOL                NewGame();
  TileState           GetTileState(UINT uiX, UINT uiY) const
                        { return m_tsGrid[uiX][uiY]; }
  BOOL                SetTileState(UINT uiX, UINT uiY);
  BOOL                GetMyTurn() const { return m_bMyTurn; }

  // Protected Methods
protected:
  BOOL                DPCreateLocalPlayer();
  BOOL                DPDestroyLocalPlayer();
  BOOL                DPSendTurnMsg(CPoint& ptPoint);
  BOOL                DPReceiveTurnMsg(CPoint& ptPoint);
  void                DPCreateEventThread();
  static UINT         DPEventMsgStart(LPVOID pData);
  void                DPEventMsg();
  BOOL                CheckLine(int i, int j, int ii, int jj);
  BOOL                IsWinner();
  BOOL                IsDraw();

  // Private Data
private:
  LPDIRECTPLAY  m_pDirectPlay;
  DPID          m_dpidPlayer[2];
  HANDLE        m_hDPEvent;
  CWinThread*   m_pEventThread;
  TileState     m_tsGrid[3][3];
  UINT          m_uiTurns;
  BOOL          m_bMyTurn;
};

#endif
